#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include "dict.h"
#include "field.h"
#include "talloc.h"

/*
 * config file syntax:
 *
 *
 * [stanza1]
 * name1:def1
 * name2:def2
 * [stanza2]
 * name1:def1
 * name3:def3
 * 
 * should produce
 * 
 * ((stanza1 ((name1 def1) (name2 def2)))
 *  (stanza2 ((name1 def1) (name3 def3)))
 * )
 */

     
/* return 1 iff 's' contains nothing but white space */

static int
blankline (char *s)
{
    while (isspace (*s))
	++s;
    return (*s == '\0');
}

/* change all NL and CR characters to SP, remove trailing SPs */

static void
cleanup (char *s)
{
    char *p = s;

    while (*p) {
	if (*p == '\n' || *p == '\r')
	    *p = ' ';
	++p;
    }

    /* now remove trailing spaces */
    while (p > s && p[-1] == ' ')
	*--p = '\0';
}

/* return the number of NL characters in 's' */

static int
countnl (char *s)
{
    int result = 0;

    while (*s)
	if (*s++ == '\n')
	    ++result;
    return result;
}

/* return a pointer to the first non-space char in 's' */

static char *
firstnonspace (char *s)
{
    while (isspace (*s))
	++s;
    return s;
}

Atom *
get_config (char *filename, FILE *fp, char *compiler)
{
    char *stanza = NULL;
    char buf[10240];
    char *ptr;
    Atom *result = dictionary ();/* root of config file parse tree */
    Atom *curdefs = NULL;	/* current definitions */
    Atom *args = NULL;
    int lineno = 1;
    int copying = 0;

    while (get_header_field (buf, sizeof (buf), fp)) {
	int nl = countnl (buf);

	cleanup (buf);
	if (blankline (buf))
	    ;
	else if (*buf == '[') {			/* [compiler] */
	    /*
	     * close existing stanza
	     */
	    if (stanza) {
		if (args)
		    curdefs = define (curdefs, ":args", args);
		args = NULL;
		result = define (result, stanza, curdefs);
		curdefs = NULL;
		free (stanza);
	    }
	    if ((ptr = strchr (buf+1, ']')) == NULL) {
		fprintf (stderr, "%s (%d): missing ']'\n",
			 filename, lineno);
		free (stanza);
		stanza = NULL;
	    }
	    else {
		*ptr = '\0';
		stanza = strdup (buf+1);
		curdefs = dictionary ();
	    }
	}
	else if (*buf == '/') {			/* /pattern/ expression */
	    if (stanza == NULL) {
		fprintf (stderr, "%s (%d): definition outside of stanza\n",
			 filename, lineno);
	    }
	    else if ((ptr = strchr (buf+1, '/')) == NULL) {
		fprintf (stderr, "%s (%d): missing '/' to end pattern\n",
			 filename, lineno);
	    }
	    else {
		*ptr = '\0';
		args = define (args, buf+1, string (firstnonspace (ptr+1)));
	    }
	    
	}
	else if (ptr = strchr (buf, ':')) {	/* name: value */
	    if (stanza == NULL) {
		fprintf (stderr, "%s (%d): definition outside of stanza\n",
			 filename, lineno);
	    }
	    else {
		*ptr = '\0';
		curdefs = define (curdefs, buf,
				  string (firstnonspace (ptr+1)));
	    }
	}
	else {
	    fprintf (stderr, "%s (%d): missing ':' in definition\n",
		     filename, lineno);
	    fprintf (stderr, "<%s>\n", buf);
	}

	lineno += nl;
    }
    if (stanza) {
	if (args)
	    curdefs = define (curdefs, ":args", args);
	args = NULL;
	result = define (result, stanza, curdefs);
	curdefs = NULL;
	free (stanza);
    }
    return result;
}

Atom *
read_config_file (char *filename)
{
    FILE *fp;
    Atom *result;

    if ((fp = fopen (filename, "r")) == NULL) {
	fprintf (stderr, "cannot open %s: %s\n", filename, strerror (errno));
	return NULL;
    }
    result = parse_config_file (filename, fp);
    fclose (fp);
    return result;
}

main ()
{
    Atom *root = parse_config_file ("stdin", stdin);
    fprint_atom (stdout, root);
    printf ("\n");
}
