#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "conf.h"

#include "args.h"
#include "cf.h"
#include "concat.h"
#include "dict.h"
#include "expr.h"
#include "global.h"
#include "httpget.h"
#include "nbindex.h"
#include "sort.h"
#include "talloc.h"
#include "type.h"
#include "unpack.h"

extern struct global global;

Atom *
sort_candidates (Atom *dict, char *key)
{
    Atom *pivot_nvp;
    Atom *value;
    int pivot_val;
    int comp_val;
    Atom *less = NULL;
    Atom *equal = NULL;
    Atom *greater = NULL;

    do {
	if (nullp (dict))
	    return NULL;
	pivot_nvp = this (dict);
	dict = next (dict);
	value = nlookup (unmeta ("candidate", next (pivot_nvp)), key);
    } while (value == NULL || stringp (value) == 0);
    
    pivot_val = atoi (stringval (value));
    equal = cons (pivot_nvp, equal);

    if (nullp (dict))
	return equal;

    while (!nullp (dict)) {
	value = nlookup (unmeta ("candidate", next (this (dict))), key);
	
	if (value == NULL || !stringp (value))
	    goto next;

	comp_val = atoi (stringval (value));
	if (comp_val < pivot_val)
	    less = cons (this (dict), less);
	else if (comp_val > pivot_val)
	    greater = cons (this (dict), greater);
	else
	    equal = cons (this (dict), equal);
    next:
	dict = next (dict);
    }
    less = sort_candidates (less, key);
    greater = sort_candidates (greater, key);
    return append (less, append (equal, greater));
}


Atom *
lookup_id (char *name, Atom *dict)
{
    Atom *result = lookup (dict, name);

    if (listp (result))
	return next (result);
    return NULL;
}

Atom *
screen_libs (Atom *list, Atom *platform_attrs)
{
    Atom *result = NULL;

    if (nullp (list))
	return NULL;

    /* screen the rest of the libs first */
    result = screen_libs (next (list), platform_attrs);

    if (listp (this (list)) && stringp (this (this (list))) &&
	strcmp (typename (next (this (list))), "candidate") == 0) {
	Atom *c;
	Atom *dict = unmeta ("candidate", next (this (list)));

	/*
	 * check version library was compiled for,
	 * against os version of target platform
	 */
	if (c = nlookup (dict, "library.target.os.version")) {
	    Atom *target_version;

	    target_version = nlookup (platform_attrs,
				      "target.os.version");
	    if (target_version &&
		version_compare (stringval (target_version),
				 stringval (c)) < 0) {
		return result;
	    }
	}

	/*
	 * now check constraint expression
	 */
	if (c = lookup (dict, "library.constraint")) {
	    c = expr_evaluate (next (c), lookup_id, platform_attrs);
	    if (integerp (c) && intval (c))
		return cons (this (list), result);
	    else
		return result;
	}

	return cons (this (list), result);
    }
    else
	return result;
}

#if 0
/*
 * given a list of libs (where each lib has associated with it a
 * list of candidates) generate a list of (lib candidate) pairings
 * that satisfy all constraints
 *
 * XXX this version doesn't support constraints yet
 * 
 * should be called as generate_clists (NULL, libs)
 *
 * where libs is
 * (("lib1" ("filename1" (name value) (name value))
 *          ("filename2" (name value) (name value)))
 *  ("lib2" ("filename3" (name value) (name value))
 *          ("filename4" (name value) (name value))
 *          ("filename5" (name value) (name value))
 *          ("filename6" (name value) (name value))))
 *
 * parent gets built up as lists of the form
 * (("lib1" ("filename1" (name value) (name value)))
 *  ("lib2" ("filename3" (name value) (name value))))
 */
     
Atom *
generate_clists (Atom *parent, Atom *libs)
{
    Atom *libname;
    Atom *cand;
    Atom *clist;
    Atom *result = NULL;

    if (nullp (libs))
	return parent;
    libname = this (this (libs));
    for (clist = next (this (libs)); !nullp (clist); clist = next (clist)) {
	cand = cons (libname, this (clist));
	result = cons (generate_clists (cons (cand, parent), next (libs)),
		       result);
    }
    return result;
}
#endif

Atom *
reverse (Atom *a)
{
    Atom *result = NULL;

    for (; !nullp (a); a = next (a))
	result = cons (this (a), result);
    return result;
}

Atom *
prepend_to_each (Atom *a, Atom *list)
{
    Atom *result = NULL;

    for (; !nullp(list); list=next(list))
	result = cons (cons (a, this (list)), result);
    return reverse (result);
}

Atom *
cartesian (Atom *list_of_lists)
{
    Atom *result = NULL;
    Atom *new_result = NULL;
    Atom *lol;

    result = NULL;
    for (lol = this (list_of_lists); !nullp (lol); lol = next (lol))
	result = cons (cons (this (lol), NULL), result);
    new_result = reverse (result);
    /* unref (result); */
    result = new_result;
#if 0
    printf ("initial result = ");
    print_atom (result);
    printf ("\n");
#endif

    for (lol = next (list_of_lists); !nullp (lol); lol = next (lol)) {
	Atom *list = this (lol);
	Atom *l;
	
#if 0
	printf ("list = ");
	print_atom (list);
	printf ("\nresult = ");
	print_atom (result);
	printf ("\n");
#endif

	new_result = NULL;
	for (l = list; !nullp (l); l = next (l)) {
	    Atom *item = this (l);
	    Atom *r;
	    
#if 0
	    printf ("prepending ");
	    print_atom (item);
	    printf (" to each element of \n");
	    print_atom (result);
	    printf ("\n");
#endif
	    for (r = result; !nullp(r); r = next(r))
		new_result = cons (cons (item, this (r)), new_result);
	}
	/* unref (result); */
	result = reverse(new_result);
    }
    return result;
}

Atom *
generate_clists (Atom *lib_list)
{
    Atom *foo;
    Atom *bar;
    Atom *ll;

    bar = NULL;
    for (ll = lib_list; !nullp (ll); ll = next (ll)) {
	Atom *lib = this (ll);
	Atom *libname = this (lib);
	Atom *inst;

	foo = NULL;
	for (inst = next (lib); !nullp (inst); inst = next (inst)) {
	    foo = cons (cons (libname, this (inst)), foo);
	}
	/* bar = cons (cons (reverse (foo), NULL), bar); */
	bar = cons (reverse (foo), bar);
    }
#if 0
    print_atom (bar);
    printf ("\n");
#endif
    return cartesian (reverse (bar));
}

#if 0
/*
 * find all candidates on netlib for a package named 'libname'
 */

static Atom *
find_netlib_pkg (Atom *libname)
{
    char *uri;
    Atom *result;
    FILE *fp;

    if (!stringp (libname))
	return NULL;

    uri = concat ("http://www.netlib.org/netbuild/packages/",
		  stringval (libname), "/", global.netbuild_arch,
		  "/netbuild.index", NULL);
    if ((fp = http_get_internal (uri, 0, NULL)) == NULL) {
	fprintf (stderr, "cannot get %s\n", uri);
	free (uri);
	return NULL;
    }
    result = parse_nb_index_file (uri, fp);
    fclose (fp);
    free (uri);
    return result;
}
#else
/*
 * find the list of netbuild packages matching 'libname' at 'baseurl'
 */

static Atom *
find_netlib_pkg_at (Atom *libname, Atom *baseurl)
{
    char *uri;
    Atom *result;
    FILE *fp;
    char *baseurlstr;

    if (!stringp (libname) || !stringp (baseurl))
	return NULL;

    /*
     * accommodate URLs either ending or not ending in '/'
     */
    baseurlstr = stringval (baseurl);
    if (baseurlstr[strlen(baseurlstr)-1] == '/')
	uri = concat (stringval (baseurl),
		      stringval (libname), "/", global.netbuild_arch,
		      "/netbuild.index", NULL);
    else
	uri = concat (stringval (baseurl), "/",
		      stringval (libname), "/", global.netbuild_arch,
		      "/netbuild.index", NULL);

    if ((fp = http_get_internal (uri, 0, NULL)) == NULL) {
	/* fprintf (stderr, "cannot get %s\n", uri); */
	free (uri);
	return NULL;
    }
    result = parse_nb_index_file (uri, fp);
    fclose (fp);
    free (uri);
    return result;
}

/*
 * find the list of netbuild packages matching 'libname' at all
 * of the base URLs listed in global.server_list
 */

static Atom *
find_netlib_pkg (Atom *libname)
{
    Atom *ptr;
    Atom *result = NULL;

    if (!listp (global.server_list)) {
	fprintf (stderr, "find_netlib_pkg: empty server list.\n");
	return NULL;
    }
    for (ptr = global.server_list; !nullp (ptr); ptr=next(ptr)) {
	Atom *this_result;

	this_result = find_netlib_pkg_at (libname, this(ptr));

	if (!nullp (this_result))
	    result = append (result, this_result);
    }
    return result;
}
#endif

/*
 * return 1 if any of the candidates in 'clist' references a local library
 * 
 * XXX dummy function for now
 */

static int
refs_local_lib (Atom *clist)
{
    return 0;
}

/*
 * find a library in a local directory
 *
 * XXX need to actually get the list of library suffixes for that
 * platform and/or compiler.
 */

static Atom *
find_local_lib (Atom *lib, Atom *dirlist)
{
    Atom *ptr;

    if (!stringp (lib))
	return NULL;
    for (ptr = dirlist; !nullp(ptr); ptr = next (ptr)) {
	Atom *dir = this (ptr);
	char *filename = concat (stringval (dir), "/lib",
				 stringval (lib), ".a", NULL);

	if (access (filename, R_OK) == 0) {
	    /* return a description of the library */
	    Atom *metadata = NULL;

	    free (filename);
	    /* local libraries have precedence 0 */
	    metadata = define (metadata, "library.precedence", string ("0"));
	    metadata = define (metadata, ":local", integer (1));

	    return cons (cons (dir, meta ("candidate", metadata)), NULL);
	}
	free (filename);
    }
    return NULL;
}

static Atom *
find_preloaded_pkg (Atom *libname)
{
    char *localdir;
    char *name;
    Atom *metadata;
    Atom *result;

    name = concat ("NB2_LIB_", stringval (ugly_encode (libname)), NULL);
    if ((localdir = getenv (name)) == NULL) {
#if 0
	printf ("$%s not found\n", name);
#endif
	free (name);
	return NULL;
    }
    free (name);

#if 0
    printf ("preloaded %s found in %s\n",
	    stringval (libname), localdir);
#endif

    if (strncmp (localdir, "file:///", 8) != 0)
	return NULL;
    metadata = read_metadata (NULL, localdir+8);
    if (nlookup (metadata, "library.precedence") == NULL)
	metadata = define (metadata, "library.precedence", string ("9999"));
    result = cons (cons (string (localdir),
			 meta ("candidate", metadata)),
		   NULL);
    return result;
}

/*
 * find eligible candidiates for a library named 'lib' -
 * either on the local disk or on the net
 *
 * XXX need to select libraries based on "mode" (static vs. dynamic)
 */

static Atom *
find_lib (Atom *lib, Atom *dirlist)
{
    Atom *result;
    Atom *libname;
    Atom *mode;

    if (listp (lib) && stringp (this (lib))) {
	mode = nlookup (next (lib), ":mode");
	libname = this (lib);
    }
    else if (stringp (lib))
	libname = lib;
    else
	return NULL;

    if (result = find_preloaded_pkg (libname))
	result = screen_libs (result, global.platform_attrs);
    else if (result = find_netlib_pkg (libname))
	result = screen_libs (result, global.platform_attrs);
    if (!refs_local_lib (result))
	result = append (find_local_lib (libname, dirlist), result);
    return reverse (sort_candidates (result, "library.precedence"));
}

/*
 * make a list of (libname candidate) matches that satisfy this
 * platform's criteria
 */

Atom *
find_libs (Atom *liblist, Atom *dirlist)
{
    Atom *lib;
    Atom *dir;
    Atom *result = NULL;
    Atom *candidates;

    for (lib = liblist; !nullp (lib); lib = next (lib)) {
	Atom *libname;

	if (listp (this (lib)))
	    libname = this (this (lib));
	else if (stringp (this (lib)))
	    libname = this (lib);

	if ((candidates = find_lib (this (lib), dirlist)) == NULL) {
	    fprintf (stderr, "find_libs: warning: cannot find library %s\n",
		     stringval (libname));
	    /* goto fail; */
	}
	else
	    result = define (result, stringval (libname), candidates);
    }
    return generate_clists (result);

 fail:
    unref (result);
    return NULL;
}

#ifdef TEST

struct global global;

main (int argc, char **argv)
{
    Atom *rlist;
    Atom *libs = NULL;
    char *uri;
    FILE *fp;
    int i;
    char *platform_name;

    platform_init (NULL);

    for (i = 1; i < argc; ++i) {
	printf ("looking up %s\n", argv[i]);
	uri = concat ("http://www.netlib.org/netbuild/packages/", argv[i], "/",
		      global.netbuild_arch, "/netbuild.index", NULL);
	if ((fp = http_get_internal (uri, 0, NULL)) == NULL) {
	    fprintf (stderr, "cannot get %s\n", uri);
	    break;
	}
#if 0
	printf ("***A %d\n", i);
	fprint_dict (stdout, global.platform_attrs);
	printf ("\n");
#endif
	rlist = parse_nb_index_file (uri, fp);
#if 0
	printf ("***B %d\n", i);
	fprint_dict (stdout, global.platform_attrs);
	printf ("\n");
#endif
	fclose (fp);
#if 0
	fprint_dict (stdout, rlist);
	printf ("\n\n");
#endif
	rlist = sort_candidates (rlist, "library.precedence");
#if 0
	fprint_dict (stdout, rlist);
	printf ("\n\n");
#endif
	libs = define (libs, argv[i],
		       screen_libs (rlist, global.platform_attrs));
#if 0
	printf ("***C %d\n", i);
	fprint_dict (stdout, global.platform_attrs);
	printf ("\n");
#endif
#if 0
	fprint_dict (stdout, libs);
	printf ("\n");
#endif
    }
    fprint_dict (stdout, libs);
    printf ("\n");
    print_atom (generate_clists (libs));
    printf ("\n");
}
#endif /* TEST */
