#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "talloc.h"
#include "type.h"
#include "ugly.h"

static int
unhex (c)
char c;
{
    if (c >= 'A' && c <= 'F')
	return c - 'A' + 10;
    if (c >= 'a' && c <= 'f')
	return c - 'a' + 10;
    if (c >= '0' && c <= '9')
	return c - '0';
}

static char
tohex (i)
unsigned int i;
{
    i = i & 0x0f;
    return i >= 10 ? (i - 10 + 'A') : (i + '0');
}

Atom *
ugly_encode (Atom *s)
{
    unsigned char *src, *dst;
    unsigned char *buf;
    int length;
    Atom *result;

    if (!stringp (s))
	return NULL;
    src = stringval (s);
    length = 0;
    while (*src) {
	if (isalnum (*src)) {
	    ++src;
	    ++length;
	}
	else {
	    ++src;
	    length += 4;
	}
    }
    dst = buf = TALLOC(length + 1, unsigned char);
    src = (unsigned char *) stringval (s);
    while (*src) {
	if (isalnum (*src))
	    *dst++ = *src++;
	else {
	    *dst++ = '_';
	    *dst++ = '_';
	    *dst++ = tohex ((*src & 0xf0) >> 4);
	    *dst++ = tohex (*src & 0x0f);
	    ++src;
	}
    }
    *dst++ = '\0';
    result = string (buf);
    free (buf);
    return result;
}


Atom *
ugly_decode (Atom *s)
{
    char *src, *dst;
    unsigned char *buf;
    int length;
    Atom *result;

    if (!stringp (s))
	return NULL;
    src = stringval (s);
    length = 0;
    while (*src) {
	if (*src == '_' && src[1] == '_' &&
	    isxdigit(src[2]) && isxdigit(src[3])) {
	    src += 4;
	    length++;
	}
	else {
	    src++;
	    length++;
	}
    }
    dst = buf = TALLOC(length + 1, unsigned char);
    src = stringval (s);
    while (*src) {
	if (*src == '_' && src[1] == '_' &&
	    isxdigit(src[2]) && isxdigit(src[3])) {
	    unsigned char byte;
	    
	    byte = (unhex (src[2]) << 4) | (unhex (src[3]));
	    *dst++ = byte;
	    src += 4;
	}
	else
	    *dst++ = *src++;
    }
    *dst++ = '\0';
    result = string (buf);
    free (buf);
    return result;
}

#ifdef TEST
main (int argc, char **argv)
{
    int i;

    for (i = 1; i < argc; ++i) {
	printf ("ugly_encode(%s) = %s\n", argv[i],
		stringval(ugly_encode(string(argv[i]))));
	printf ("ugly_decode(%s) = %s\n", argv[i],
		stringval(ugly_decode(string(argv[i]))));
	printf ("ugly_encode(ugly_decode(%s)) = %s\n",
		argv[i],
		stringval(ugly_encode(ugly_decode(string(argv[i])))));
	printf ("ugly_decode(ugly_encode(%s)) = %s\n",
		argv[i],
		stringval(ugly_decode(ugly_encode(string(argv[i])))));
    }
}
#endif
