#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include "concat.h"
#include "global.h"
#include "httpget.h"
#include "pathnames.h"
#include "tar.h"
#include "tarlib.h"
#include "verbosity.h"

extern struct global global;

static int
cb (struct tar_mem *memhdr, void *closure)
{
    int i;
    int verbosity = (int) closure;

    if (memhdr->name[0] == '/' ||
	strncmp (memhdr->name, "../", 3) == 0 ||
	strncmp (memhdr->name + strlen (memhdr->name) - 3, "/..", 3) == 0 ||
	strstr (memhdr->name, "/../") != NULL) {
	fprintf (stderr,
		 "unpack: component %s references nonlocal directory\n",
		 memhdr->name);
	return 0;
    }
    if (verbosity >= VERBOSITY_LIB_FILES)
	printf ("[extracting %s]\n", memhdr->name);
    return 1;
}

/*
 * download a package containing files needed to set up nb2,
 * and extract those files.
 *
 * we always exit immediately after returning, so there's no
 * need to clean up memory or file descriptors.
 *
 * XXX really need a way to do integrity checking, to make
 * sure we download the entire file intact.
 */

int
nb_setup ()
{
    FILE *in, *out;
    char tempfilename[1024];
    char *url = "http://www.netlib.org/netbuild/setup/nb2-setup.tar";
    char *command;
    char buf[8*1024];
    int nread;
    char *home;
    int tar_fd;

    if (getenv ("NB_SERVER_URL"))
	url = concat (getenv ("NB_SERVER_URL"), "/../setup/nb2-setup.tar", 
                      NULL);

    fprintf (stderr, "nb2: downloading setup files...\n");

    sprintf (tempfilename, "/tmp/%d.tar", getpid ());
    if ((out = fopen (tempfilename, "w")) == NULL) {
	perror (tempfilename);
	goto fail;
    }
    if ((in = http_get_internal (url, 0, NULL)) == NULL) {
	fprintf (stderr, "nb2: download failed:\n %s\n", url);
	goto fail;
    }
    while ((nread = fread (buf, sizeof (char), sizeof (buf), in)) > 0) {
	if (fwrite (buf, sizeof (char), nread, out) != nread)
	    goto fail;
    }
    fclose (in);
    fclose (out);

    if ((home = getenv ("HOME")) == NULL) {
	fprintf (stderr, "no HOME environment variable\n");
	goto fail;
    }

    fprintf (stderr, "nb2: installing setup files in %s...\n",
	     global.path_prefix);

    if (chdir (home) < 0) {
	perror (concat ("chdir (", home, ")", NULL));
	goto fail;
    }
    mkdir ("NetBuild2", 0711);
    if (chdir ("NetBuild2") < 0) {
	perror (concat ("chdir(", home, "/NetBuild2)", NULL));
	goto fail;
    }

#if 0
    /*
     * XXX calling tar makes me nervous
     * 1. tar might not be in the user's PATH
     * 2. syntax might vary slightly from one system to another
     * 3. there's no way to keep tar from scribbling on arbitrary
     *    directories or files (if, for example, the netlib server
     *    got compromised or an evil proxy got in the way)
     * solution: provide our own tar decoder.
     */
    command = concat ("tar xvfp ", tempfilename, NULL);
    if (system (command) != 0) {
	fprintf (stderr, "nb2: command failed: %s\n", command);
	goto fail;
    }
#else
    tar_fd = open (tempfilename, O_RDONLY);
    if (tar_rd (tar_fd, global.path_prefix, cb, (void *) 1) == NULL)
	fprintf (stderr, "error extracting setup files\n");
    close (tar_fd);
#endif
    unlink (tempfilename);
    return 0;
    
 fail:
    unlink (tempfilename);
    return 1;
}
